@@ -1,4 +1,4 @@
-/* $Header: /u/christos/cvsroot/tcsh/ed.term.h,v 1.13 1996/09/24 16:57:20 christos Exp $ */
+/* $Header: /p/tcsh/cvsroot/tcsh/ed.term.h,v 1.19 2015/03/25 19:53:16 christos Exp $ */
 /*
  * ed.term.h: Local terminal header
  */
@@ -14,11 +14,7 @@
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice, this list of conditions and the following disclaimer in the
  *    documentation and/or other materials provided with the distribution.
- * 3. All advertising materials mentioning features or use of this software
- *    must display the following acknowledgement:
- *	This product includes software developed by the University of
- *	California, Berkeley and its contributors.
- * 4. Neither the name of the University nor the names of its contributors
+ * 3. Neither the name of the University nor the names of its contributors
  *    may be used to endorse or promote products derived from this software
  *    without specific prior written permission.
  *
@@ -304,10 +300,18 @@
 # endif /* masscomp */
 #endif /* CBRK */
 #ifndef CMIN
-# define CMIN		CEOF
+# if VMIN == VEOF
+#  define CMIN		CEOF
+# else
+#  define CMIN		1
+# endif
 #endif /* CMIN */
 #ifndef CTIME
-# define CTIME		CEOL
+# if VTIME == VEOL
+#  define CTIME		CEOL
+# else
+#  define CTIME		0
+# endif
 #endif /* CTIME */
 
 /*
@@ -454,9 +458,10 @@
 /*
  * fix for hpux10 inconsistency: it has VWERASE, but TIOCSLTC returns
  * EINVAL if one tries to change it
+ * Also for RH6.2 on the alpha, defined TIOCGLTC, but does not have
+ * struct ltchars
  */
-#if defined(hpux) && defined(VSUSP) && defined(VDSUSP) && defined(VWERASE) && d
-efined(VLNEXT)
+#if (defined(hpux) && defined(VSUSP) && defined(VDSUSP) && defined(VWERASE) && defined(VLNEXT)) || defined(__linux__) || defined(__GNU__) || defined(__GLIBC__) || defined(__QNXNTO__)
 # undef TIOCGLTC       /* not really needed */
 # undef TIOCSLTC
 #endif
@@ -489,4 +494,37 @@ efined(VLNEXT)
 #define C_NCC		25
 #define C_SH(A)		(1 << (A))
 
+/*
+ * Terminal dependend data structures
+ */
+typedef struct {
+#ifdef WINNT_NATIVE
+    int dummy;
+#else /* !WINNT_NATIVE */
+# if defined(POSIX) || defined(TERMIO)
+#  ifdef POSIX
+    struct termios d_t;
+#  else
+    struct termio d_t;
+#  endif /* POSIX */
+# else /* SGTTY */
+#  ifdef TIOCGETP
+    struct sgttyb d_t;
+#  endif /* TIOCGETP */
+#  ifdef TIOCGETC
+    struct tchars d_tc;
+#  endif /* TIOCGETC */
+#  ifdef TIOCGPAGE
+    struct ttypagestat d_pc;
+#  endif /* TIOCGPAGE */
+#  ifdef TIOCLGET
+    int d_lb;
+#  endif /* TIOCLGET */
+# endif /* POSIX || TERMIO */
+# ifdef TIOCGLTC
+    struct ltchars d_ltc;
+# endif /* TIOCGLTC */
+#endif /* WINNT_NATIVE */
+} ttydata_t;
+
 #endif /* _h_ed_term */
