@@ -1,4 +1,4 @@
-/* $Header: /p/tcsh/cvsroot/tcsh/sh.glob.c,v 3.80 2010/05/20 22:03:40 christos Exp $ */
+/* $Header: /p/tcsh/cvsroot/tcsh/sh.glob.c,v 3.89 2015/05/05 14:27:20 christos Exp $ */
 /*
  * sh.glob.c: Regular expression expansion
  */
@@ -32,7 +32,7 @@
  */
 #include "sh.h"
 
-RCSID("$tcsh: sh.glob.c,v 3.80 2010/05/20 22:03:40 christos Exp $")
+RCSID("$tcsh: sh.glob.c,v 3.89 2015/05/05 14:27:20 christos Exp $")
 
 #include "tc.h"
 #include "tw.h"
@@ -516,11 +516,11 @@ globone(Char *str, int action)
 	stderror(ERR_NAME | ERR_NOMATCH);
     }
  result:
-    if (vl[0] == NULL) {
+    if (vl && vl[0] == NULL) {
 	xfree(vl);
 	return (Strsave(STRNULL));
     }
-    if (vl[1]) 
+    if (vl && vl[1]) 
 	return (handleone(str, vl, action));
     else {
 	str = strip(*vl);
@@ -693,7 +693,7 @@ dobackp(Char *cp, int literal)
 static void
 backeval(struct blk_buf *bb, struct Strbuf *word, Char *cp, int literal)
 {
-    int icnt;
+    ssize_t icnt;
     Char c, *ip;
     struct command faket;
     int    hadnl;
@@ -781,13 +781,12 @@ backeval(struct blk_buf *bb, struct Strbuf *word, Char *cp, int literal)
 		stderror(ERR_OLD);
 	    alias(&paraml);
 	    t = syntax(paraml.next, &paraml, 0);
-	    if (t == NULL)
-		return;
 	    cleanup_push(t, syntax_cleanup);
 	    /* The F_BACKQ flag must set so the job output is correct if
 	     * printexitvalue is set.  If it's not set, the job output
 	     * will have "Exit N" appended where N is the exit status. */
-	    t->t_dflg = F_BACKQ|F_NOFORK;
+	    if (t)
+		    t->t_dflg = F_BACKQ|F_NOFORK;
 	    if (seterr)
 		stderror(ERR_OLD);
 #ifdef SIGTSTP
@@ -808,7 +807,7 @@ backeval(struct blk_buf *bb, struct Strbuf *word, Char *cp, int literal)
     c = 0;
     ip = NULL;
     do {
-	int     cnt = 0;
+	ssize_t     cnt = 0;
 	char   *tmp;
 
 	tmp = tibuf;
@@ -881,7 +880,7 @@ backeval(struct blk_buf *bb, struct Strbuf *word, Char *cp, int literal)
 	if (c != 0 && (cnt || literal))
 	    pword(bb, word);
 	hadnl = 0;
-    } while (c > 0);
+    } while (c >= 0);
  eof:
     cleanup_until(&pvec[0]);
     pwait();
