@@ -1,4 +1,4 @@
-/* $Header: /src/pub/tcsh/tc.printf.c,v 3.25 2004/08/04 17:12:31 christos Exp $ */
+/* $Header: /p/tcsh/cvsroot/tcsh/tc.printf.c,v 3.37 2014/02/28 17:10:31 christos Exp $ */
 /*
  * tc.printf.c: A public-domain, minimal printf/sprintf routine that prints
  *	       through the putchar() routine.  Feel free to use for
@@ -34,51 +34,50 @@
  */
 #include "sh.h"
 
-RCSID("$Id: tc.printf.c,v 3.25 2004/08/04 17:12:31 christos Exp $")
+RCSID("$tcsh: tc.printf.c,v 3.37 2014/02/28 17:10:31 christos Exp $")
 
 #ifdef lint
 #undef va_arg
 #define va_arg(a, b) (a ? (b) 0 : (b) 0)
 #endif
 
-#define INF	32766		/* should be bigger than any field to print */
+#define INF	INT_MAX		/* should be bigger than any field to print */
 
-static char buf[128];
 static char snil[] = "(nil)";
 
-static	void	xaddchar	__P((int));
-static	void	doprnt		__P((void (*) __P((int)), const char *, va_list));
+static	void	xaddchar	(int);
+static	int	doprnt		(void (*) (int), const char *, va_list);
 
-static void
-doprnt(addchar, sfmt, ap)
-    void    (*addchar) __P((int));
-    const char   *sfmt;
-    va_list ap;
+static int
+doprnt(void (*addchar) (int), const char *sfmt, va_list ap)
 {
     char *bp;
     const char *f;
 #ifdef SHORT_STRINGS
-    Char *Bp;
+    const Char *Bp;
 #endif /* SHORT_STRINGS */
-#ifdef HAVE_QUAD
+#ifdef HAVE_LONG_LONG
     long long l;
     unsigned long long u;
 #else
     long l;
     unsigned long u;
 #endif
+    char buf[(CHAR_BIT * sizeof (l) + 2) / 3 + 1]; /* Octal: 3 bits per char */
     int i;
     int fmt;
     unsigned char pad = ' ';
-    int     flush_left = 0, f_width = 0, prec = INF, hash = 0, do_long = 0;
-    int     sign = 0;
+    int     flush_left = 0, f_width = 0, prec = INF, hash = 0;
+    int	    do_long = 0, do_size_t = 0, do_ptrdiff_t = 0;
+    int     sign = 0, count = 0;
     int     attributes = 0;
 
 
     f = sfmt;
     for (; *f; f++) {
 	if (*f != '%') {	/* then just out the char */
-	    (*addchar) ((int) (((unsigned char)*f) | attributes));
+	    (*addchar) (((unsigned char)*f) | attributes);
+	    count++;
 	}
 	else {
 	    f++;		/* skip the % */
@@ -129,6 +128,14 @@ doprnt(addchar, sfmt, ap)
 		    f++;
 		}
 	    }
+	    if (*f == 'z') {	/* size_t format */
+		do_size_t++;
+		f++;
+	    }
+	    if (*f == 't') {	/* ptrdiff_t format */
+		do_ptrdiff_t++;
+		f++;
+	    }
 
 	    fmt = (unsigned char) *f;
 	    if (fmt != 'S' && fmt != 'Q' && isupper(fmt)) {
@@ -140,15 +147,18 @@ doprnt(addchar, sfmt, ap)
 	    case 'd':
 		switch (do_long) {
 		case 0:
-		    l = (long) (va_arg(ap, int));
+		    if (do_size_t)
+			l = (long) (va_arg(ap, size_t));
+		    else
+			l = (long) (va_arg(ap, int));
 		    break;
 		case 1:
-#ifndef HAVE_QUAD
+#ifndef HAVE_LONG_LONG
 		default:
 #endif
 		    l = va_arg(ap, long);
 		    break;
-#ifdef HAVE_QUAD
+#ifdef HAVE_LONG_LONG
 		default:
 		    l = va_arg(ap, long long);
 		    break;
@@ -166,13 +176,19 @@ doprnt(addchar, sfmt, ap)
 		    *bp++ = '-';
 		f_width = f_width - (int) (bp - buf);
 		if (!flush_left)
-		    while (f_width-- > 0) 
-			(*addchar) ((int) (pad | attributes));
-		for (bp--; bp >= buf; bp--) 
-		    (*addchar) ((int) (((unsigned char) *bp) | attributes));
+		    while (f_width-- > 0)  {
+			(*addchar) (pad | attributes);
+			count++;
+		    }
+		for (bp--; bp >= buf; bp--)  {
+		    (*addchar) (((unsigned char) *bp) | attributes);
+		    count++;
+		}
 		if (flush_left)
-		    while (f_width-- > 0)
-			(*addchar) ((int) (' ' | attributes));
+		    while (f_width-- > 0) {
+			(*addchar) (' ' | attributes);
+			count++;
+		    }
 		break;
 
 	    case 'p':
@@ -185,15 +201,20 @@ doprnt(addchar, sfmt, ap)
 	    case 'u':
 		switch (do_long) {
 		case 0:
-		    u = (unsigned long) (va_arg(ap, unsigned int));
+		    if (do_size_t)
+			u = va_arg(ap, size_t);
+		    else if (do_ptrdiff_t)
+			u = va_arg(ap, ptrdiff_t);
+		    else
+			u = va_arg(ap, unsigned int);
 		    break;
 		case 1:
-#ifndef HAVE_QUAD
+#ifndef HAVE_LONG_LONG
 		default:
 #endif
 		    u = va_arg(ap, unsigned long);
 		    break;
-#ifdef HAVE_QUAD
+#ifdef HAVE_LONG_LONG
 		default:
 		    u = va_arg(ap, unsigned long long);
 		    break;
@@ -226,19 +247,24 @@ doprnt(addchar, sfmt, ap)
 		}
 		i = f_width - (int) (bp - buf);
 		if (!flush_left)
-		    while (i-- > 0)
-			(*addchar) ((int) (pad | attributes));
+		    while (i-- > 0) {
+			(*addchar) (pad | attributes);
+			count++;
+		    }
 		for (bp--; bp >= buf; bp--)
-		    (*addchar) ((int) (((unsigned char) *bp) | attributes));
+		    (*addchar) (((unsigned char) *bp) | attributes);
 		if (flush_left)
-		    while (i-- > 0)
-			(*addchar) ((int) (' ' | attributes));
+		    while (i-- > 0) {
+			(*addchar) (' ' | attributes);
+			count++;
+		    }
 		break;
 
 
 	    case 'c':
 		i = va_arg(ap, int);
-		(*addchar) ((int) (i | attributes));
+		(*addchar) (i | attributes);
+		count++;
 		break;
 
 	    case 'S':
@@ -251,29 +277,31 @@ doprnt(addchar, sfmt, ap)
 	        }
 		f_width = f_width - Strlen(Bp);
 		if (!flush_left)
-		    while (f_width-- > 0)
+		    while (f_width-- > 0) {
 			(*addchar) ((int) (pad | attributes));
+			count++;
+		    }
 		for (i = 0; *Bp && i < prec; i++) {
-#ifdef WIDE_STRINGS
 		    char cbuf[MB_LEN_MAX];
 		    size_t pos, len;
-#endif
 
-		    if (fmt == 'Q' && *Bp & QUOTE)
-			(*addchar) ((int) ('\\' | attributes));
-#ifdef WIDE_STRINGS
+		    if (fmt == 'Q' && *Bp & QUOTE) {
+			(*addchar) ('\\' | attributes);
+			count++;
+		    }
 		    len = one_wctomb(cbuf, *Bp & CHAR);
-		    for (pos = 0; pos < len; pos++)
-			(*addchar) ((int) ((unsigned char)cbuf[pos]
-					   | attributes | (*Bp & ATTRIBUTES)));
-#else
-		    (*addchar) ((int) ((*Bp & TRIM) | attributes));
-#endif
+		    for (pos = 0; pos < len; pos++) {
+			(*addchar) ((unsigned char)cbuf[pos] | attributes
+				    | (*Bp & ATTRIBUTES));
+			count++;
+		    }
 		    Bp++;
 		}
 		if (flush_left)
-		    while (f_width-- > 0)
-			(*addchar) ((int) (' ' | attributes));
+		    while (f_width-- > 0) {
+			(*addchar) (' ' | attributes);
+			count++;
+		    }
 		break;
 #endif /* SHORT_STRINGS */
 
@@ -283,20 +311,26 @@ doprnt(addchar, sfmt, ap)
 lcase_s:
 		if (!bp)
 		    bp = snil;
-		f_width = f_width - strlen((char *) bp);
+		f_width = f_width - strlen(bp);
 		if (!flush_left)
-		    while (f_width-- > 0)
-			(*addchar) ((int) (pad | attributes));
+		    while (f_width-- > 0) {
+			(*addchar) (pad | attributes);
+			count++;
+		    }
 		for (i = 0; *bp && i < prec; i++) {
-		    if (fmt == 'q' && *bp & QUOTE)
-			(*addchar) ((int) ('\\' | attributes));
-		    (*addchar) ((int) (((unsigned char) *bp & TRIM) |
-				   	attributes));
+		    if (fmt == 'q' && *bp & QUOTE) {
+			(*addchar) ('\\' | attributes);
+			count++;
+		    }
+		    (*addchar) (((unsigned char) *bp & TRIM) | attributes);
+		    count++;
 		    bp++;
 		}
 		if (flush_left)
-		    while (f_width-- > 0)
-			(*addchar) ((int) (' ' | attributes));
+		    while (f_width-- > 0) {
+			(*addchar) (' ' | attributes);
+			count++;
+		    }
 		break;
 
 	    case 'a':
@@ -304,24 +338,26 @@ lcase_s:
 		break;
 
 	    case '%':
-		(*addchar) ((int) ('%' | attributes));
+		(*addchar) ('%' | attributes);
+		count++;
 		break;
 
 	    default:
 		break;
 	    }
-	    flush_left = 0, f_width = 0, prec = INF, hash = 0, do_long = 0;
+	    flush_left = 0, f_width = 0, prec = INF, hash = 0;
+	    do_ptrdiff_t = 0, do_size_t = 0, do_long = 0;
 	    sign = 0;
 	    pad = ' ';
 	}
     }
+    return count;
 }
 
 
 static char *xstring, *xestring;
 static void
-xaddchar(c)
-    int     c;
+xaddchar(int c)
 {
     if (xestring == xstring)
 	*xstring = '\0';
@@ -330,91 +366,87 @@ xaddchar(c)
 }
 
 
-pret_t
+int
 /*VARARGS*/
-#ifdef FUNCPROTO
 xsnprintf(char *str, size_t size, const char *fmt, ...)
-#else
-xsnprintf(va_alist)
-    va_dcl
-#endif
 {
+    int count;
     va_list va;
-#ifdef FUNCPROTO
     va_start(va, fmt);
-#else
-    char *str, *fmt;
-    size_t size;
-
-    va_start(va);
-    str = va_arg(va, char *);
-    size = va_arg(va, size_t);
-    fmt = va_arg(va, char *);
-#endif
 
     xstring = str;
     xestring = str + size - 1;
-    doprnt(xaddchar, fmt, va);
+    count = doprnt(xaddchar, fmt, va);
     va_end(va);
     *xstring++ = '\0';
-#ifdef PURIFY
-    return 1;
-#endif
+    return count;
 }
 
-pret_t
+int
 /*VARARGS*/
-#ifdef FUNCPROTO
 xprintf(const char *fmt, ...)
-#else
-xprintf(va_alist)
-    va_dcl
-#endif
 {
+    int count;
     va_list va;
-#ifdef FUNCPROTO
     va_start(va, fmt);
-#else
-    char   *fmt;
-
-    va_start(va);
-    fmt = va_arg(va, char *);
-#endif
-    doprnt(xputchar, fmt, va);
+    count = doprnt(xputchar, fmt, va);
     va_end(va);
-#ifdef PURIFY
-    return 1;
-#endif
+    return count;
 }
 
-
-pret_t
-xvprintf(fmt, va)
-    const char   *fmt;
-    va_list va;
+int
+xvprintf(const char *fmt, va_list va)
 {
-    doprnt(xputchar, fmt, va);
-#ifdef PURIFY
-    return 1;
-#endif
+    return doprnt(xputchar, fmt, va);
 }
 
-pret_t
-xvsnprintf(str, size, fmt, va)
-    char   *str;
-    size_t size;
-    const char   *fmt;
-    va_list va;
+int
+xvsnprintf(char *str, size_t size, const char *fmt, va_list va)
 {
+    int count;
     xstring = str;
     xestring = str + size - 1;
-    doprnt(xaddchar, fmt, va);
+    count = doprnt(xaddchar, fmt, va);
     *xstring++ = '\0';
-#ifdef PURIFY
-    return 1;
-#endif
+    return count;
 }
 
+char *
+xvasprintf(const char *fmt, va_list va)
+{
+    size_t size;
+    char *buf;
+
+    buf = NULL;
+    size = 2048; /* Arbitrary */
+    for (;;) {
+	va_list copy;
+
+	buf = xrealloc(buf, size);
+	xstring = buf;
+	xestring = buf + size - 1;
+	va_copy(copy, va);
+	doprnt(xaddchar, fmt, copy);
+	va_end(copy);
+	if (xstring < xestring)
+	    break;
+	size *= 2;
+    }
+    *xstring++ = '\0';
+    return xrealloc(buf, xstring - buf);
+}
+
+char *
+xasprintf(const char *fmt, ...)
+{
+    va_list va;
+    char *ret;
+
+    va_start (va, fmt);
+    ret = xvasprintf(fmt, va);
+    va_end(va);
+    return ret;
+}
 
 
 #ifdef PURIFY
@@ -429,37 +461,20 @@ xvsnprintf(str, size, fmt, va)
 #define FILE int
 #endif
 int 
-#ifdef FUNCPROTO
 fprintf(FILE *fp, const char* fmt, ...)
-#else
-fprintf(va_alist)
-    va_dcl
-#endif
 {
+    int count;
     va_list va;
-#ifdef FUNCPROTO
     va_start(va, fmt);
-#else
-    FILE *fp;
-    const char   *fmt;
-
-    va_start(va);
-    fp = va_arg(va, FILE *);
-    fmt = va_arg(va, const char *);
-#endif
-    doprnt(xputchar, fmt, va);
+    count = doprnt(xputchar, fmt, va);
     va_end(va);
-    return 1;
+    return count;
 }
 
 int 
-vfprintf(fp, fmt, va)
-    FILE *fp;
-    const char   *fmt;
-    va_list va;
+vfprintf(FILE *fp, const char *fmt, va_list va)
 {
-    doprnt(xputchar, fmt, va);
-    return 1;
+    return doprnt(xputchar, fmt, va);
 }
 
 #endif	/* PURIFY */
